@@ -16,6 +16,7 @@ import importlib
 import warnings
 
 from keystoneauth1 import adapter
+import keystoneauth1.exceptions.catalog
 from keystoneauth1 import plugin
 from keystoneauth1 import session
 import requestsexceptions
@@ -70,7 +71,7 @@ def _make_key(key, service_type):
 class CloudConfig(object):
     def __init__(self, name, region, config,
                  force_ipv4=False, auth_plugin=None,
-                 openstack_config=None):
+                 openstack_config=None, session_constructor=None):
         self.name = name
         self.region = region
         self.config = config
@@ -79,6 +80,7 @@ class CloudConfig(object):
         self._auth = auth_plugin
         self._openstack_config = openstack_config
         self._keystone_session = None
+        self._session_constructor = session_constructor or session.Session
 
     def __getattr__(self, key):
         """Return arbitrary attributes."""
@@ -101,6 +103,10 @@ class CloudConfig(object):
     def __ne__(self, other):
         return not self == other
 
+    def set_session_constructor(self, session_constructor):
+        """Sets the Session constructor."""
+        self._session_constructor = session_constructor
+
     def get_requests_verify_args(self):
         """Return the verify and cert values for the requests library."""
         if self.config['verify'] and self.config['cacert']:
@@ -154,9 +160,12 @@ class CloudConfig(object):
         # atrocity from the as-yet-unsullied eyes of our users.
         # Of course, if the user requests a volumev2, that structure should
         # still work.
-        if (service_type == 'volume' and
-                self.get_api_version(service_type).startswith('2')):
-            service_type = 'volumev2'
+        # What's even more amazing is that they did it AGAIN with cinder v3
+        if service_type == 'volume':
+            if self.get_api_version(service_type).startswith('2'):
+                service_type = 'volumev2'
+            elif self.get_api_version(service_type).startswith('3'):
+                service_type = 'volumev3'
         return self.config.get(key, service_type)
 
     def get_service_name(self, service_type):
@@ -196,7 +205,7 @@ class CloudConfig(object):
                     " since verify=False".format(
                         cloud=self.name, region=self.region))
             requestsexceptions.squelch_warnings(insecure_requests=not verify)
-            self._keystone_session = session.Session(
+            self._keystone_session = self._session_constructor(
                 auth=self._auth,
                 verify=verify,
                 cert=cert,
@@ -246,11 +255,18 @@ class CloudConfig(object):
         if service_key == 'identity':
             endpoint = session.get_endpoint(interface=plugin.AUTH_INTERFACE)
         else:
-            endpoint = session.get_endpoint(
-                service_type=self.get_service_type(service_key),
-                service_name=self.get_service_name(service_key),
-                interface=self.get_interface(service_key),
-                region_name=self.region)
+            args = {
+                'service_type': self.get_service_type(service_key),
+                'service_name': self.get_service_name(service_key),
+                'interface': self.get_interface(service_key),
+                'region_name': self.region
+            }
+            try:
+                endpoint = session.get_endpoint(**args)
+            except keystoneauth1.exceptions.catalog.EndpointNotFound:
+                self.log.warning("Keystone catalog entry not found (%s)",
+                                 args)
+                endpoint = None
         return endpoint
 
     def get_legacy_client(
@@ -351,26 +367,64 @@ class CloudConfig(object):
                 constructor_kwargs['api_version'] = version
             else:
                 constructor_kwargs['version'] = version
+        if service_key == 'database':
+            # TODO(mordred) Remove when https://review.openstack.org/314032
+            # has landed and released. We're passing in a Session, but the
+            # trove Client object has username and password as required
+            # args
+            constructor_kwargs['username'] = None
+            constructor_kwargs['password'] = None
 
         return client_class(**constructor_kwargs)
 
     def _get_swift_client(self, client_class, **kwargs):
+        auth_args = self.get_auth_args()
+        auth_version = self.get_api_version('identity')
         session = self.get_session()
         token = session.get_token()
         endpoint = self.get_session_endpoint(service_key='object-store')
         if not endpoint:
             return None
+        # If we have a username/password, we want to pass them to
+        # swift - because otherwise it will not re-up tokens appropriately
+        # However, if we only have non-password auth, then get a token
+        # and pass it in
         swift_kwargs = dict(
+            auth_version=auth_version,
             preauthurl=endpoint,
             preauthtoken=token,
-            auth_version=self.get_api_version('identity'),
             os_options=dict(
+                region_name=self.get_region_name(),
                 auth_token=token,
                 object_storage_url=endpoint,
-                region_name=self.get_region_name()),
-        )
+                service_type=self.get_service_type('object-store'),
+                endpoint_type=self.get_interface('object-store'),
+
+            ))
         if self.config['api_timeout'] is not None:
             swift_kwargs['timeout'] = float(self.config['api_timeout'])
+
+        # create with password
+        swift_kwargs['user'] = auth_args.get('username')
+        swift_kwargs['key'] = auth_args.get('password')
+        swift_kwargs['authurl'] = auth_args.get('auth_url')
+        os_options = {}
+        if auth_version == '2.0':
+            os_options['tenant_name'] = auth_args.get('project_name')
+            os_options['tenant_id'] = auth_args.get('project_id')
+        else:
+            os_options['project_name'] = auth_args.get('project_name')
+            os_options['project_id'] = auth_args.get('project_id')
+
+        for key in (
+                'user_id',
+                'project_domain_id',
+                'project_domain_name',
+                'user_domain_id',
+                'user_domain_name'):
+            os_options[key] = auth_args.get(key)
+        swift_kwargs['os_options'].update(os_options)
+
         return client_class(**swift_kwargs)
 
     def get_cache_expiration_time(self):
@@ -407,3 +461,53 @@ class CloudConfig(object):
             if resource not in expiration:
                 return default
             return float(expiration[resource])
+
+    def get_external_networks(self):
+        """Get list of network names for external networks."""
+        return [
+            net['name'] for net in self.config['networks']
+            if net['routes_externally']]
+
+    def get_external_ipv4_networks(self):
+        """Get list of network names for external IPv4 networks."""
+        return [
+            net['name'] for net in self.config['networks']
+            if net['routes_ipv4_externally']]
+
+    def get_external_ipv6_networks(self):
+        """Get list of network names for external IPv6 networks."""
+        return [
+            net['name'] for net in self.config['networks']
+            if net['routes_ipv6_externally']]
+
+    def get_internal_networks(self):
+        """Get list of network names for internal networks."""
+        return [
+            net['name'] for net in self.config['networks']
+            if not net['routes_externally']]
+
+    def get_internal_ipv4_networks(self):
+        """Get list of network names for internal IPv4 networks."""
+        return [
+            net['name'] for net in self.config['networks']
+            if not net['routes_ipv4_externally']]
+
+    def get_internal_ipv6_networks(self):
+        """Get list of network names for internal IPv6 networks."""
+        return [
+            net['name'] for net in self.config['networks']
+            if not net['routes_ipv6_externally']]
+
+    def get_default_network(self):
+        """Get network used for default interactions."""
+        for net in self.config['networks']:
+            if net['default_interface']:
+                return net['name']
+        return None
+
+    def get_nat_destination(self):
+        """Get network used for NAT destination."""
+        for net in self.config['networks']:
+            if net['nat_destination']:
+                return net['name']
+        return None
